/*
 * AutoSnail.java
 *
 * Created on 01 February 2005, 14:46
 */

package MiscTests;

/**
 *
 * @author  Jean-Alain Grunchec
 */
import java.awt.*;
import java.awt.event.*;
import java.applet.*;
import java.text.NumberFormat;
import java.net.*;

/** The main application class.
 * 
 *  This class is responsible for displaying all of the GUI, apart
 *  from the map display.  It also responds to button clicks and map
 *  selection events.
 */
class AutoSnail extends DemoFrame implements SelectionListener {//, DistanceWatcher
    private Map map = null;
        // The map, loaded from a text file
    private MapViewer viewer = new MapViewer();
        // A view of the map; we listen for selection events
    private PathFinder pathfinder;
        // Thread for the search, if one is in progress
    private Town from, to;
        // Source and destination towns, if selected
    private int selStage = 0;
        // Number of towns chosen so far; when it reaches 2, we're
        // ready to search.
    private Thread searcher = null;
    
    // The control panel (right-hand column) with various text fields
    // and check boxes.
    private Panel controls = 
        new Panel(new ColumnLayout(5, 10, 2, ColumnLayout.LEFT));
    private Font textFont = new Font("SansSerif", Font.PLAIN, 12);
    private Font labelFont = new Font("SansSerif", Font.BOLD, 12);
    private PassiveTextField fromText = new PassiveTextField(20);
    private PassiveTextField toText = new PassiveTextField(20);
    private PassiveTextField distText = new PassiveTextField(20);
    private Checkbox workCheck = new Checkbox("Show working");
    private Checkbox heurCheck = new Checkbox("Use heuristic");
    private Checkbox delayCheck = new Checkbox("Delay between steps");
    
    // The scroll bar for delay between steps
    private Scrollbar delaySlider = 
        new Scrollbar(Scrollbar.HORIZONTAL, 500, 100, 0, 1000) {
            // Make it quite small
            public Dimension getPreferredSize() {
                return new Dimension(100, 15);
            }
        };

    // The buttons at the bottom
    private Panel buttons = new Panel(new FlowLayout());
    
    /** Make a button, connect it to the listener, and put it in the panel */
    private Button makeButton(String text, ActionListener listener) {
        Button button = new Button(text);
        button.addActionListener(listener);
        buttons.add(button);
        return button;
    }
    
    private Button searchButton = makeButton("Search", new ActionListener() {
            public void actionPerformed(ActionEvent e) { search(); }
        });
    private Button resetButton = makeButton("Reset", new ActionListener() {
            public void actionPerformed(ActionEvent e) { reset(); }
        });
    private Button quitButton = makeButton("Quit", new ActionListener() {
            public void actionPerformed(ActionEvent e) { die(); }
        });

    /** Make a label with specified text and font */
    private Label makeLabel(String text, Font font) {
        Label label = new Label(text);
        label.setFont(font);
        return label;
    }
    
    public AutoSnail(boolean isApplet) {
        super("AutoSnail Express", isApplet);
        this.setFont(textFont);

        // Basic layout: map with controls at right, buttons at bottom
        setLayout(new BorderLayout());
        this.add(viewer, "Center");
        this.add(controls, "East");
        this.add(buttons, "South");
        controls.setBackground(Color.lightGray);
        buttons.setBackground(Color.lightGray);
        viewer.setBackground(Color.white);
        viewer.setSelectionListener(this);

        controls.add(makeLabel("From:", labelFont));
        controls.add(fromText);
        controls.add(makeLabel("To:", labelFont));
        controls.add(toText);
        controls.add(makeLabel("Distance:", labelFont));
        controls.add(distText);
        controls.add(workCheck);
        controls.add(heurCheck);
        controls.add(delayCheck);
        controls.add(delaySlider);
        
        // Checking or clearing the delay box calls our setDelay() method
        delayCheck.addItemListener(new ItemListener() {
            public void itemStateChanged(ItemEvent e) {
                delaySlider.setVisible(delayCheck.getState());
                setDelay();
            }
        });
        
        // Moving the delay slider also calls setDelay()
        delaySlider.addAdjustmentListener(new AdjustmentListener() {
            public void adjustmentValueChanged(AdjustmentEvent e) {
                setDelay();
            }
        });
        
        delaySlider.setBlockIncrement(100);
        delaySlider.setUnitIncrement(100);

        setLocation(250, 50);
        pack();

        // The delay slider is invisible until the check box is selected
        delaySlider.setVisible(false);
    }
    
    /** Respond to selection events on towns */
    public void select(Town t) {
        switch (selStage) {
        case 0:
            // Select first town
            from = t;
            from.setColor(Color.green);
            viewer.updateTown(from);
            fromText.setText(from.getName());
            selStage = 1;
            break;
            
        case 1:
            // Select second town
            if (from == t) break;
            to = t;
            to.setColor(Color.red);
            viewer.updateTown(to);
            toText.setText(to.getName());
            searchButton.setEnabled(true);
            selStage = 2;
            break;
                
        default:
            // Ignore selections after first two
            break;
        }
    }
    
    /** Adjust the delay to match what is selected by the controls */
    private void setDelay() {
        if (pathfinder == null) return;
        
        if (delayCheck.getState())
            pathfinder.setDelay(delaySlider.getValue());
        else
            pathfinder.setDelay(0);
    }
            
    /** If all is ready, carry out a search */
    public void search() { 
        if (selStage < 2) return;
        searchButton.setEnabled(false);
        resetButton.setEnabled(false);
        resetSearch();
        viewer.repaint();
        
        boolean heur = heurCheck.getState();
        boolean work = workCheck.getState();
        pathfinder = new PathFinder(map, from, to, heur, work);
        setDelay();
        pathfinder.setViewer(viewer);
      //  pathfinder.setWatcher(this);
        searcher = new PFThread(pathfinder);
        searcher.start();
    }
    
    /** A PFThread is a concurrent thread that carries out a search
     * 
     *  The concurrency is needed in order to allow display updating
     *  during the search: otherwise, the search becomes a
     *  subroutine of the event loop, and all display updates are
     *  delayed until the end of the search.
     */
    private class PFThread extends Thread {
        PathFinder pf;
        
        PFThread(PathFinder pf) {
            this.pf = pf;
        }
        
        public void run() {
            float dist = pf.search();
            done(dist);
        }
    }
    
    private static NumberFormat fmt = NumberFormat.getInstance();
    static { fmt.setMaximumFractionDigits(1); }                                                 
    
    /** Update distance display to show progress */
    public synchronized void update(float dist) {
        distText.setText(fmt.format(dist));
    }
    
    /** Update display when search is over */
    private void done(float dist) {
        String s;
        if (dist < PathFinder.INF)
            s = fmt.format(dist);
        else
            s = "infinity";
        
        distText.setText(s);
        searchButton.setEnabled(true);
        resetButton.setEnabled(true);
    }
    
    /** Reset the selection of towns and the search */
    private void reset() { 
        resetSearch();
        if (from != null) from.setColor(Color.black);
        if (to != null) to.setColor(Color.black);
        searchButton.setEnabled(false);
        selStage = 0;
        fromText.setText("");
        toText.setText("");
        controls.repaint();
        viewer.repaint();
    }
    
    /** Prepare for a new search with the same towns */
    private void resetSearch() {
        map.reset();
        if (from != null) from.setColor(Color.green);
        if (to != null) to.setColor(Color.red);
        distText.setText("");
    }
    
    public void setMap(Map map) {
        this.map = map;
        viewer.setMap(map);
        reset();
    }

    public void initApplet(DemoApplet applet) {
        super.initApplet(applet);
        
        try {
            URL url = new URL(applet.getDocumentBase(), applet.getParameter("map"));
            setMap(Map.fromURL(url));
        } 
        catch (MalformedURLException e) { }
    }
    
    public void die() {
        if (searcher != null) searcher.stop();
        super.die();
    }

    public static void main(String args[]) {
        AutoSnail frame = new AutoSnail(false);
        frame.setMap(Map.fromFile("C:\\Zip_05_01\\src\\MiscTests\\mapinfo.txt"));
        frame.show();
    }
}

/** An uneditable text field with a white background */
class PassiveTextField extends TextField {
    public PassiveTextField(int width) {
        super(width);
        setEditable(false);
        setBackground(Color.white);
    }
}


