/*
 * Town.java
 *
 * Created on 01 February 2005, 14:44
 */

package MiscTests;

/**
 *
 * @author  Jean-Alain Grunchec
 */
/* Copyright (C) 2002 J. M. Spivey */

import java.util.*;
import java.awt.Color;

/** All permanent data about a town.
 * 
 * This class represents towns with a name, a location, and a list of
 * outgoing arcs.  In addition, the town has a colour that is visible when
 * it is displayed, and an object 'data' associated with it.  In Dijkstra's
 * algorithm, the data includes the status, estimated distance and backlink.
 */
class Town {
    private String name;
    private Coords location;
    private Vector arcs = new Vector(5);
    private Color color = Color.black;
    private Object data;
    
    public Town(String name, Coords location) {
        this.name = name;
        this.location = location;
    }
    
    public void addArc(Town dest, Road road) {
        arcs.addElement(new Arc(dest, road));
    }
    
    public String getName() { return name; }
    public Coords getLocation() { return location; }
    public Enumeration getArcs() { return arcs.elements(); }
    
    public Color getColor() { return color; }
    public void setColor(Color color) { this.color = color; }
    
    public float distance(Town other) {
        return this.location.distance(other.location);
    }

    public void setData(Object data) { this.data = data; }
    public Object getData() { return data; }
}

/** A symmetrical road between two towns.
 * 
 * The road has a length, and a colour that is used when it is displayed.  Each road is
 * associated with two directed Arcs that are used in Dijkstra's algorithm.
 */
class Road {
    public Town t1, t2;
    public float length;
    private Color color = Color.black;
    
    public Road(Town t1, Town t2, float length) {
        this.t1 = t1;
        this.t2 = t2;
        this.length = length;
    }
    
    public Color getColor() { return color; }
    public void setColor(Color color) { this.color = color; }
}

/** A directed link from one town to another.
 * 
 * Each arc is associated with a road, but has a direction, so that one of the
 * road's two ends is identified as its destination.
 */
class Arc {
    private Town dest;
    private Road road;
    
    public Arc(Town dest, Road road) {
        this.dest = dest;
        this.road = road;
    }
    
    public Town getDest() { return dest; }
    public Road getRoad() { return road; }
}

/** Coordinates for a point on the map.
 * 
 * The coordinates are floating-point values expressed in miles.
 * They must be transformed to integer coordinates expressed in
 * pixels before the point can be displayed.
 */
class Coords {
    private float x, y;
    
    public Coords(float x, float y) {
        this.x = x;
        this.y = y;
    }
    
    public Coords() {
        this(0.0f, 0.0f);
    }
    
    public float getX() { return x; }
    public float getY() { return y; }
    
    public float distance(Coords other) {
        float dx = this.x - other.x, dy = this.y - other.y;
        return (float) Math.sqrt(dx*dx + dy*dy);
    }
}
