package guikit; 

import neuroml.core.*;
import java.lang.reflect.*;

import neuroml.sim.ElementChangedListener;


/** Represents a field of an object
 */
public class ObjectNode {  
    Object parent = null;   // Owning class
    Object content;  // the value
    Field  field = null;

    DisplayHints displayHints = null; 

    /** True if parent is a list
     */
    boolean inlist=false;
    String parentPath=""; // The path of this object's parent
    String path=""; // The path of this object, from the root of the namespace

    /** Notify this if the object content changes
     */
    ElementChangedListener ecl;

    /** Example use:
     * new ObjectNode( testObject, "field1" );
     */
    public ObjectNode(Object content) {
	this.content = content;
    }  

    /** Constructor to use for elements of a list/set
     */
    public ObjectNode(Object list, Object content, ElementChangedListener ecl) {
	this.parent = list;
	this.content = content;
	this.ecl = ecl;
	inlist = true;

    }  

    /** Constructor to use for fields of a parent object 
     */
    public ObjectNode(Object parent, String fieldName, ElementChangedListener ecl) {
	this(parent,fieldName);
	this.ecl = ecl;
    }
    public ObjectNode(Object parent, String fieldName) {
	this.parent = parent;
	if (parent != null) 
	    try {
		Class c = parent.getClass();
		field = c.getField(fieldName);
		content = field.get(parent);
		//System.out.println("Object Node constructor class "+c+" field "+field+" content "+content);
	    } catch (Exception e) {
		System.out.println("Object Node error "+e);	   
	    }    
    }
    
    /** Any display hints
     * Note: display hints can be added using static <fieldname>Hints = new DisplayHints(0,123); 
     */
    public void setDisplayHints( DisplayHints h ) { displayHints =h; }
    public DisplayHints getDisplayHints() { return displayHints; }

    /** The "path" of this object, i.e. root in current namespace from top
     */
    public void setPath(String p) { path=p; }

    /** Set path to be parent+"."+field (or "field" if parent empty)
     */
    public void setPath(String parent, String field) { 
	parentPath = parent;
	if (parent.equals("")) path=""+field; 
	else path = parent+"."+field;
    }

    /** Return the path of this object
     */
    public String getPath() { return path; }

    /** Return the path of this object's parent
     * NOTE this isn't updated if the parent name somewhere changes :(
     */
    public String getParentPath() { return parentPath; }

    public String getName() { 
	if (content==null) { return "null"; }
	return (field==null) ? content.getClass().getName(): field.getName();    
    }

    /** Alter this node's name */
    public void changeFieldName( String n ) {
	setPath(parentPath, n);
    }
    
    public Object getContent() { 
	return content; 
    }

    /** Update the content of this node */
    public void setContent(Object c) { 
	//System.out.println("setContent "+c+ " parent "+parent+" field "+field);
	if ((parent!=null) && !inlist) {
	    try { 
		field.set(parent,c); 

	    } catch (Exception e ){
	
		System.out.println("INFORMATION (not a bug) setContent Exception "+e+" Content "+c+ " parent "+parent+" field "+field);
		//System.out.println("class "+c.getClass()+" superclass "+c.getClass().getSuperclass()+" package "+c.getClass().getPackage());
	    } 
	    if (ecl!=null) ecl.elementChanged(content, c);
	} else if (inlist) {
	    if (ecl!=null) ecl.elementChanged(content, c);
	}
	content = c; 	
    }

    public Field getField() { return field; }
    public Object getParent() { return parent; }

    public String toString(){
	String title = getName()+" : "+content;
	if (content instanceof DValue) {
	    DValue dvObj = (DValue)content;
	    title = getName()+" : "+dvObj.value+" Unit : "+dvObj.unit.getTargetName();
	}
	else if (content instanceof Ref) {
	    Ref refObj = (Ref)content;
	    title = getName()+" : "+refObj.getTargetName();
	}
	return title;
    }
}
