/** MPI based SimKernel


The basic algorithm is:
     rank 0 runs script reader, issueing commands
     other ranks run a slave, waiting for commands, then running them.

Commands are distributed using two MPI_BCasts; the first gives the
size of the command in the second.

At present, only MPI_COMM_WORLD is used.

The last command is a Poison command which terminates the kernels.

 */

package neosim.mpikernel;

import neosim.kernel.*;
import neosim.iface.*;
import neosim.implementation.*;
 
import mpi.* ;
import java.io.*;
 

import java.util.Vector;

import java.util.Random;
import java.util.Hashtable;
import java.util.Enumeration;
 

/** Basic test command */
class TestCommand extends CommandObj {
    String msg;
    public TestCommand(int id, String msg) {
        super(id);
        this.msg = msg;
    }
    public void doit(SimKernel sk) {
        System.out.println("Executing a test command " + msg);
    }
}

class Poison extends CommandObj {
    public Poison(int id) { super(id); }
    public void doit(SimKernel sk) {
        sk.terminate();
    }
}



/** MPI Simulation Kernel
 */
public class MpiKernel extends BasicKernel {
    
    boolean poisoned = false;
    double runt = 0.0;
    String args[];
  
  
    public MpiKernel(String args[]) {
        super();
        this.args = args;
    }
 
    /** Kill all threads... */
    public void terminate() {
        poisoned = true;
    }

    /** Barrier  */
    public void barrier() {
        MPI.COMM_WORLD.Barrier() ;
    }
         
    public int rank;

    /** Send next command */
    public void sendCommand( CommandObj mo ) {
        try {
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            ObjectOutputStream oos = new ObjectOutputStream(baos);
            oos.writeObject(mo);
            oos.flush();            
            byte[] dataOut = baos.toByteArray();
            int size[] = new int[1];
            size[0] = dataOut.length;
            MPI.COMM_WORLD.Bcast(size, 0, 1, MPI.INT, 0);
            MPI.COMM_WORLD.Bcast(dataOut, 0, size[0], MPI.BYTE, 0);
        } catch (Exception e) { System.out.println(e); }
        mo.doit(this);
    }
    
    /** Send command synchronously */
    public void syncCommand( CommandObj mo ) {
        sendCommand(mo);
    }
    

    /** Recv next command */
    CommandObj getCommand() {
        ByteArrayInputStream bais;
        ObjectInputStream ois;
        int size[] = new int[1];
        MPI.COMM_WORLD.Bcast(size, 0, 1, MPI.INT, 0);
        byte raw[] = new byte[size[0]];
        MPI.COMM_WORLD.Bcast(raw, 0, size[0], MPI.BYTE, 0);
        CommandObj mo = null;
        try {
            bais = new ByteArrayInputStream(raw);
            ois  = new ObjectInputStream(bais);
            mo = (CommandObj) ois.readObject();
        } catch (Exception e) { System.out.println(e); }
        return mo;
    }

    /** This runs on root */
    public void scriptReader() {
        System.out.println("Starting script reader");
        sendCommand(new TestCommand(1,"Hello"));
        sendCommand(new TestCommand(2,"from"));
        sendCommand(new Poison(3));
        System.out.println("Done script reader");
    }

    /** This runs everywhere else */
    public void slave() {    
        System.out.println("Starting slave");
        CommandObj co;
        do {
            co = getCommand();
            co.doit(this);
        } while (!poisoned);
        System.out.println("Done slave");
    }


    public void init() {
        rank = MPI.COMM_WORLD.Rank();
        System.out.println("["+rank+"] About to do kernel init."+args.length);
        for (int i=0; i<args.length; i++) { 
            System.out.println(i+ " "+args[i]);
        }
        System.out.println("==== About to barrier.");
        MPI.COMM_WORLD.Barrier();
        System.out.println("==== Done barrier.");

        //      MPI.Init(args);

        System.out.println("Done kernel init.");


        int myrank = MPI.COMM_WORLD.Rank() ;
        int mysize = MPI.COMM_WORLD.Size() ;
        System.out.println("["+myrank+"] kernel init done, "+mysize+" in total.");
        if (rank==0) { scriptReader(); }
        else { slave(); }

        //      MPI.Finalize();
    }

   
    /** Extra commands from SimKernel */
    public void reset() { System.out.println("Reset not yet implemented\n"); }
}

