package neosim.implementation;
import neosim.iface.*;
import java.util.Vector;
import java.util.Enumeration;

/**
 * This class implements an event queue used internally  to
 * manage
 * the list of future and deferred Events.   It works like a normal FIFO
 * queue, but during insertion events are kept in order from the smallest time
 * stamp to the largest. This means the next event to occur will be at the top
 * of the queue. <P>
 * The current implementation
 * is uses a Vector to store the queue and is inefficient for popping
 * and inserting elements because the rest of the array has to be
 * moved down one space. A better method would be to use a circular array.
 * @version     0.1, 15 June 1999
 * @author      Fred Howell
 */

public class EventListImpl extends Vector implements EventList {

    /**
     * Allocates a new EventListImpl object.
     */
    public EventListImpl() {
	super();
    }
    
    /**
     * Allocates a new EventList object, with an initial capacity.
     * @param initialCapacity	The initial capacity of the queue.
     */
    public EventListImpl(int initialCapacity) {
	super(initialCapacity);
    }
    
    /**
     * Remove and return the event at the top of the queue.
     * @return           The next event.
     */
    public synchronized final Event pop() {
        Event event;
            event = (Event)firstElement();
            removeElementAt(0);
	return event;
    }
    
    /**
     * Return the event at the top of the queue, without removing it.
     * @return	The next event.
     */
    public synchronized final Event top() {
	return (Event)firstElement();
    }
    
    /**
     * Remove all elements
     */
    public synchronized final void clear() {
            super.clear();
            super.setSize(0);
    }
    
    /**
     * Add a new event to the queue, preserving the temporal order of the
     * events in the queue.
     * @param new_event	The event to be put on the queue.
     */
    public final synchronized void add(Event new_event) {
	int i;
	Enumeration e;
	//	System.out.println("Adding an event to the queue : "+new_event);
	i = -1;
	for (e = elements(); e.hasMoreElements() && (i == -1);) {
	    Event event = (Event)e.nextElement();
	    if (event.getTime().t > new_event.getTime().t)
		i = indexOf(event);
	}
	
	if (i == -1)
	    addElement((Object)new_event);
	else
	    insertElementAt((Object)new_event, i);
	//	System.out.println("ok.");
    }
    
    /**
     * Returns true if there are any more events in the list
     * @return true if any more events
     */
    public synchronized final boolean hasMoreElements() {
	if (size()>0) return true; else return false;
    }   

    /** Return the vector of events */
    public synchronized final Vector getEvents() { return this; }

}
