/* Entqueue.java
 */

package neosim.kernel;

import neosim.iface.*;

import java.util.*;

/**
 * This class implements a queue of <em>entities</em>. 
 * It works like a normal FIFO
 * queue, but during insertion entities are kept in order from the 
 * smallest right window time
 * stamp to the largest. This means the next entity to be activated
 * will be at the top of the queue. <P>
 * 
 * @see		neosim.SimSystem
 * @version     0.2, 28 Sept 2000
 * @author      Fred Howell
 */

public class Entqueue   {
    TreeSet ts;
    boolean debug = false;

    /**
     * Allocates a new Entqueue object.
     */
    public Entqueue() {
	ts = new TreeSet();
    }


    /**
     * Remove and return the event at the top of the queue.
     * @return           The next event.
     */
    public final synchronized Entity pop() {
 	if (ts.isEmpty()) return null; 
	else {
	    Entity ent = (Entity)ts.first();
	    ts.remove(ent);
	    if (debug) System.out.println("Entqueue Popping an entity "+ent.getName());
	    return ent;
	}
    }
    
    /**
     * Return the event at the top of the queue, without removing it.
     * @return	The next event.
     */
    public final synchronized Entity top() {
	return (Entity)ts.first();
    }

    /**
     * @return	 the number of entities on the queue
     */
    public final synchronized int size() {
	return ts.size();
    }


    /**
     * Calculate the minimum right window time.
     * @return	The min entity right winow time.
     */
    public final synchronized Time getMinrw() {
 	if (ts.isEmpty()) {
	    Time minrw = new Time(); minrw.t=-1;
	    return minrw;
	}	
	Entity e = (Entity)ts.first();

	if (e.getID() == 0) {
	    //	    System.out.println("**** Entqueue.getMinrw() returning Sys Ent time");
	}
	return e.getRightWindowTime();
    }
    

    /**
     * Add a new entity to the queue, preserving the temporal order of the
     * right windows in the queue.
     * @param new_ent	The entity to be put on the queue.
     */
    public final synchronized void add(Entity new_ent) {
	if (debug)  System.out.println("Adding an entity to Entqueue "+new_ent.getName());
	ts.add( new_ent );
    }
}
