/** ProcManager.java
 * Maintains the status of each processor.
 */

package neosim.kernel;

import java.util.Vector;
import java.util.Enumeration;
import neosim.iface.Time;

/**
 * This class implements a vector of SimProcessors.
 *
 * Operations:<p>
 *  SetRunning - flags processor as running.<p>
 *  SetStopped - flags processor as stopped, notifies any<p>
 *		 blocked processors.<p>
 *  Wakemeup   - calls processor.pauseme, makes sure it gets<p>
 * 		 restarted the next time a processor does a Remove.<p>
 *  Getmintime - Returns earliest safe time of running processors (or -1)<p>
 *		 if none running.<p>
 * @see		neosim.kernel.SimSystem, neosim.kernel.SimProcessor
 * @version     0.1, 7 July 1998
 *              0.2, 2 Nov 1998 - changed to vector of RmtProcs.
 * @author      Fred Howell
 */

public class ProcManager extends Vector implements RmtProcMgr {
 //   neosim.kernel.DeadlockTracker tracker=new neosim.kernel.DeadlockTracker();
    Time min_time = new Time(0);
    int nrunning  = 0;
    int nodeid    = -1;
    int numnodes  = -1;
    Vector nodes;
    
    boolean debug = false;
    
    //JAG DANGEROUS OPTIMIZATION
    boolean calculated=false;
    double increment;
    /**
     * Allocates a new ProcManager object.
     */
    public ProcManager() {
	super();
    }
    
    /**
     * Allocates a new ProcManager object, with an initial capacity.
     * @param initialCapacity	The initial capacity of the queue.
     */
    public ProcManager(int initialCapacity) {
	super(initialCapacity);
    }

 
    /**
     * Return processor i.
     */  
    public final RmtProc getProcessor(int i) {
	return (RmtProc)elementAt(i);
    }
    
    /**
     * Set node id
     */  
    public final void setNodeid(int i) {
	nodeid = i;
    }
    
    /**
     * Set nodes vector (vector of RmtApp)
     */  
    public final void setNodes(Vector n) {
	nodes = n;
    }

    /**
     * Set nodes vector (vector of RmtApp)
     */  
    public final Vector getNodes() {
	return nodes;
    }
    
    /**
     * Set num nodes
     */  
    public final void setNumNodes(int i) {
	numnodes = i;
    }

    /** Return the minimum time of all running entities
     * @return The minimum time, or -1.0 if no running entities.
     */
    public synchronized final Time getMinTime() {
	return min_time;
    }
    
    /** Set the minimum time of all running entities (test)
     */
    public final synchronized void setMinTime(Time t) { 
	min_time = t;
    }
    
    /**
     * Wake this thread up when any processor finishes.
     * @param proc	The processor to send to sleep.
     */
    public final void setSleeping(RmtProc proc) {
	try {
	    if (debug)	   System.out.println(""+proc.getID()+"P SLEEPING");
	    proc.setState( SimProcessor.SLEEPING );
	    proc.pauseme(); 
	} catch (Exception e) {
	    System.out.println("SetSleeping exception "+e);
	}
    }
    
    /**
     * Flag a processor as running
     * @param proc	The processor about to start running
     */
    public final synchronized void setRunning(RmtProc proc, Time rw) {
	try {
	    if (debug)	   System.out.println(""+proc.getID()+"P RUNNING");
	    proc.setState( SimProcessor.RUNNING );
	    proc.setRightWindowTime(rw);
            //neosim.kernel.Deadlock_Message ecp5=new neosim.kernel.Deadlock_Message();
            //neosim.kernel.DeadlockTracker.LockDeadlockTracker(rw,ecp5);            
	    calcMinTime();
            //neosim.kernel.DeadlockTracker.UnLockDeadlockTracker(ecp5);
	    nrunning ++;
	} catch (Exception e) {
	    System.out.println("SetRunning exception "+e);
	}
    }
    
    
    /**
     * Update current minimum safe time.
     * @param prw 	The processor's right window
     */
    public final synchronized void updateMinTime(Time prw) {
	if (min_time.t < prw.t) min_time = prw;
    }
    
    
    /**
     * Flag a processor as just finished running.
     * @param proc	The processor about to stop running
     */
    public final synchronized void setStopped(RmtProc proc) {
	try {
	    if (debug)  System.out.println(""+proc.getID()+" stopped");
	    proc.setState( SimProcessor.STOPPED );
	    nrunning --; 
            //neosim.kernel.Deadlock_Message ecp5=new neosim.kernel.Deadlock_Message();
            //neosim.kernel.DeadlockTracker.LockDeadlockTracker(proc,ecp5);                
	    calcMinTime();
            //neosim.kernel.DeadlockTracker.UnLockDeadlockTracker(ecp5);
	    // Wake up sleepers *** I think this isn't thread safe! ***
	    for (int i=0; i<size(); i++) {
		RmtProc p = getProcessor(i);
		if (p.getState_() == SimProcessor.SLEEPING) {
		    p.restart();
		} 
	    }
	    if (debug)  System.out.println(""+proc.getID()+" stopped: new min time:"+min_time.s());
	} catch (Exception e) {
	    System.out.println("SetStopped exception "+e);
	}
    }
    
    /**
     * Calc min time of running processors
     */
    public final synchronized void calcMinTime() {
	//	nrunning --; // What is this doing here???
        
        
        
        if(!calculated){//JAG DANGEROUS OPTIMIZATION
            Time new_min_time = new Time(); new_min_time.t = -1;
            for (int i=0; i<size(); i++) {
                try {
               
                    RmtProc p = getProcessor(i);
                //System.out.println(p.getClass().getName()+" "+p.getID());
                    if (p.getState_() == SimProcessor.RUNNING) {
                        if (new_min_time.t==-1) {new_min_time = p.getRightWindowTime();}
                        else if (p.getRightWindowTime().t < new_min_time.t) {
                            new_min_time = p.getRightWindowTime();
                        }
                    }
                } catch (Exception e) {
                    System.out.println("calc min time exception "+e);
                }
            }
            min_time = new_min_time; 
            //JAG DANGEROUS OPTIMIZATION
            calculated=true;
            increment=new_min_time.t;
        }
        else
            min_time.t+=increment;
    }
    
    
}
