package neuroml.model.channel;

import neuroml.core.*;

/** 
 * Describes a single shell model for a concentration pool.  With a total
 * outward ionic current I from all current sources for the ion, the ionic
 * concentration C obeys an equation <BR>
 * dC/dt = -B*I - (C - restingConcen)/tau <BR>
 * with the initial value of C = restingConcen.  The restingConcen value
 * (resting level of the concentration) is obtained from the value of the
 * initialConcentration field of solution.
 * @author David Beeman, University of Colorado, Boulder
 */
public class ConcenPool extends LoggedObject {

    /** Constant of proportionality for the rate of increase of
      * concentration with incoming current.  In principle <BR>
      * B =  1/(ion_charge * Faraday * shell_volume) <BR>
      * In practice, B is a parameter to be fitted or estimated from
      * experiment, as buffering, non-uniform distribution of the ion, etc.,
      * will modify this value.
      */
    public DValue B = new DValue();

    /** Time constant of decay of the concentration. */
    public DValue tau = new DValue();

    /** Thickness of the shell used for the concentration pool. */
    public DValue thickness = new DValue();

    /** Contains the initialConcentration (restingConcen), ionName, and
        valency of the ion. */
    public Solution solution;

    /** A set of objects (normally of class CurrentSource), providing the
        sources of ionic currents used to calculate I. */
    public Set currentSources = new Set("CurrentSource");

    public ConcenPool() {
    }

    public ConcenPool(String name) {
	super(name);
    }

}
