package neuroml.util;

import java.net.*;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.zip.*;
import java.util.jar.*;
import java.io.*;

//JAG
import java.util.Vector;
import neuroml.NeuromlPkg;

/**
 * A class loader for loading jar files, both local and remote.
 */
public class JarClassLoader extends URLClassLoader {
    
    Hashtable<String,URL> urlsIn = new Hashtable<String,URL>();
    private boolean debug = false;

    /**
     * Creates a new JarClassLoader for the specified url.
     *
     * @param url the url of the jar file
     */
    public JarClassLoader() {
	super(new URL[] { });
    }

    public JarClassLoader(URL url) {
	super(new URL[] { url },ClassLoader.getSystemClassLoader());
    }

    public JarClassLoader(URL[] urls) {
	super(urls,ClassLoader.getSystemClassLoader());
    }
    
    /** Check if url already registered
     */
    private boolean alreadyIn(String jarurl){
	int slash = jarurl.lastIndexOf('/');// are urls always composed of / ?
	String key = jarurl.substring(slash+1);
	return (urlsIn.containsKey(key));
    }
    
    /** Registers url in hashtable
     */
    private void registerUrl(URL url){
	int slash = url.toString().lastIndexOf('/');// are urls always composed of / ?
	String key = url.toString().substring(slash+1);
	urlsIn.put(key,url);
    }

    /** Add new url to classloader path
     */
    public void add(URL url) {
	if (!alreadyIn(url.toString())){
	    super.addURL(url);
	    registerUrl(url);
	}	
    }

    /** Add new url from url's array to classloader path
     */
    public void add(URL[] urls) {
	for (int i=0;i<urls.length;i++){
	    if (!alreadyIn(urls[i].toString())){
		super.addURL(urls[i]);
		registerUrl(urls[i]);
	    }
	}
    }

    /** Add new url from string to classloader path
     */
    public void add(String jarurlname) {
	URL jarurl = null;
	
	try {
	    if ( (jarurlname.startsWith("file:")) ||
		 (jarurlname.startsWith("http:")) ||
		 (jarurlname.startsWith("ftp:"))
		 )
		jarurl = new URL(jarurlname);
	    else
		jarurl = new URL("file:"+jarurlname);	    		  
	} catch (MalformedURLException eurl) {
	    System.out.println("JarClassLoader Malformed URL "+eurl);
	}
	if (jarurl != null){
	    if (!alreadyIn(jarurl.toString())){
		super.addURL(jarurl);
		registerUrl(jarurl);
	    }
	}
    }

    public Class loadClassFromJar(String name) {
	Class c = null;
	c = getLoadedClass(name);
	if (c!=null)
	    return c;
	try {
	    c = findClass(name); 
	} catch (ClassNotFoundException e) {
	    if (debug) System.err.println("JarClassLoader loadClassFromJar Exception "+e); 
	} catch (NullPointerException ne){
	    if (debug) System.err.println("JarClassLoader loadClassFromJar Null Exception "+ne);
	}
	return c;
    }
 

    public Class loadMainClassFromJar(URL jarurl) {
	// find main class name in manifest
	
	// add and register url if not already in
	//resetClassLoader();
	add(jarurl);

	boolean empty = true;
	Manifest mf = null;
	byte buffer[] = new byte[1024];
	ZipInputStream zis = null;
	try {

	    InputStream inf = jarurl.openStream();
	    zis = new ZipInputStream(inf);
	    ZipEntry ent = null;

	    while ((ent = zis.getNextEntry()) != null) {
	        empty = false;

		String name = ent.getName();
		String type = new String();

		/* the object we're loading */
		ByteArrayOutputStream baos = new ByteArrayOutputStream();

		/* NOTE: We don't know the size of an entry until
		   we've reached the end of one because of
		   compression. This means we can't just do a get size
		   and read in the entry. */

		for (;;) {
		    int len = zis.read(buffer);
		    if (len < 0) {
			break;
		    }
		    baos.write(buffer, 0, len);
		}

		byte[] buf = baos.toByteArray();
		int size = buf.length;

		if (Manifest.isManifestName(name)) {
		    type = "manifest/manifest";
		}

	        if (type.equals("manifest/manifest")) {
		    mf = new Manifest(buf);
		}
	    }

	} catch (IOException ex) {
	    System.out.println("IOException loading archive: " + ex);
	} catch (Throwable th) {
	    System.out.println("Caught " + th + " in loadit()");
	} finally {
	    if (zis != null) {
		try {
		    zis.close();
		} catch (Exception ex) {
		    // ignore
		}
	    }
	}

	// Unfortunately ZipInputStream doesn't throw an exception if you hand
	// it a non-Zip file.  Our only way of spotting an invalid Jar file
	// is if there no ZIP entries.
	if (empty) {
	    System.out.println("jar file empty");

	    return null;
	}

	// find class name


	if (mf==null)
	    return null;
	String res = null;
	//	Vector manifestHeaders = mf.;
	for (Enumeration enum_ = mf.entries() ; enum_.hasMoreElements() ;) {
	    MessageHeader mh = (MessageHeader)enum_.nextElement();


	    res = mh.findValue("Main-Class");
	    if (res==null) res = mh.findValue("Name");
	    if (res!=null) break;
	}
	 

	if (res==null)
	    return null;

	String name = res;
	Class c = null;

	c = getLoadedClass(name);
	if (c!=null)
	    return c;

	try {
	    c = findClass(name);
	} catch (ClassNotFoundException e) {
	    if (debug) System.err.println("JarClassLoader loadClassFromJar Exception "+e);
	} catch (NullPointerException ne){
	    if (debug) System.err.println("JarClassLoader loadClassFromJar Null Exception "+ne);
	} catch (LinkageError le){
	    if (debug) System.err.println("JarClassLoader loadClassFromJar Linkage Error "+le);
	}
	return c;
    }


    /** Try to load class using parent's class loader
     */
    public Class getLoadedClass(String classname){
	//return super.findLoadedClass(classname);
	Class c = null;
	try {
	    c = super.loadClass(classname);
	} catch (ClassNotFoundException e){ 
	    if (debug) System.err.println("JarClassLoader getLoadedClass Exception "+e); 
	}
	return c;
    }

    public void resolvesClass(Class theClass){
	try {
	   super.resolveClass(theClass);
	} catch (Exception e){
	    if (debug) System.err.println("JarClassLoader resolvesClass Exception "+e);
	}
    }
        //JAG
    public String[] readJar(String _jarPath) throws Exception{  
        try{    
            JarFile jFile=new JarFile(new URL(_jarPath).getFile());    
            Enumeration jarEntries=jFile.entries();    
            Vector<String> ListClass=new Vector<String>(1,1);
             
            while(jarEntries.hasMoreElements()){      
                ZipEntry entry=(ZipEntry)jarEntries.nextElement();      
                if(!entry.isDirectory()){        
                    String name=entry.getName().replace('/','.');        
                    if(name.endsWith(".class")){          
                        name=name.substring(0,name.length()-".class".length());   
                        //System.out.println(name);
                        Class cls=null;          
                        try{         
                            //cls=Class.forName(name);
                            cls=super.loadClass(name);
                            //Object O=cls.newInstance();
                            //O.finalize()
                            ListClass.add(name);
                        }catch(Exception e){          
                        }         
                        if(cls!=null){            //do something with the class cls          
                        }        
                    }        
                } //while
                
            }  
            return NeuromlPkg.vectorToArray(ListClass);
        }catch(Exception e)
        {  
            return null;
        }       
    }

}

