package neuroml.util;

import neuroml.core.*;
import neuroml.model.channel.*;
import neuroml.model.cell.*;
import neuroml.model.network.*;

import java.lang.reflect.*;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Collection;
import java.util.Map;
import java.util.StringTokenizer;
import java.beans.*;
import java.lang.Integer;
import java.net.*;

import java.io.*;
import org.jdom.*;
  
/**
 * Call an XML parser on a NeuroML file.
 * Uses JDOM on the tree document created in order to fulfill the NeuroML specific data structure
 * generic process, the data structure is not entirely known -> use of reflection methods
 *
 * @author Paul Rogister      
 *
 * @version 1.0, 02 March   2001  
 * @version 1.1, 09 March   2001  Fred Howell - added extra facilities for extensions
 * @version 1.2, 25 July    2001  Fred Howell - move references to NeuromlSys
 * @version 1.3, 05 August  2001  Paul Rogister - move to new NeuroML version
 * @version 1.4, 30 Jan     2002  Fred Howell - Added option for <ClassName> not <setfield class=/>
 * @version 1.5, 8 Feb      2002  Fred Howell - Renamed as NeuroMLParser
 * @version 1.6, 12 Apr     2002  Fred Howell - Fixed DValue parsing problem
 *                 
 */
public class NeuroMLParser {
    Document domDocument;
    org.jdom.Element root;
    Object desc; // The top level object
    Namespace ns;
    String cwd;

    private boolean debug = false;    
    private String nmlschema=null;

    /** Return the nml schema (optional nmlschema= tag at the top of an xml document, 
     */
    public String getNmlSchema() { return nmlschema; }

    /** Constructs a parser for the given document with references
     * put into the given namespace
     */
    public NeuroMLParser(Document document, String namespace) {
	domDocument = document;	
	ns = NeuromlSys.getNamespace( namespace );
	if (ns==null) {
	    ns = NeuromlSys.makeNamespace(namespace); // Put refs into new namespace
	}

	// Get the root URL
	cwd="";
	int l = namespace.lastIndexOf("/");
	if (l!=-1) {
	    cwd = namespace.substring( 0, l );
	}

	root = domDocument.getRootElement(); 

	// Check for a nmlschema= and <modules> sections and add jars to classpath
	Vector<URL> jarPaths = extractURL(root);
	
	// Add these URLs to the classLoader path
	// Note - if any of these urls are to NeuromlSchema this won't work yet...
	// We'll need to fetch, & compile them (maybe cache them into a moduleCache...)
	if (NeuromlSys.classLoader==null) {
	    NeuromlSys.classLoader = new JarClassLoader((URL[])jarPaths.toArray(new URL[0]));
	} else {
	    NeuromlSys.classLoader.add((URL[])jarPaths.toArray(new URL[0]));
	}

	desc = startElement(root,null); // null : no parent
	NeuromlSys.addReferences(desc,namespace);
    }

    /** Returns namespace of this document
     */
    public Namespace getNamespace() { return ns; }

    /**
     * Converts string to int 
     *              
     * @param  s String to be converted      
     * @return int conversion of String          
     */
    protected static int atoi(String s) { return (Integer.valueOf(s)).intValue(); }
    
    protected static short atos(String s) { return (Short.valueOf(s)).shortValue(); }

    /**
     * Converts string to double
     *              
     * @param  s String to be converted      
     * @return double conversion of String          
     */
    protected static double atod(String s) { return (Double.valueOf(s)).doubleValue(); }

    /**
     * Converts string to float
     *              
     * @param  s String to be converted      
     * @return float conversion of String          
     */
    protected static float atof(String s) { return (Float.valueOf(s)).floatValue(); }

    /**
     * Converts string to boolean
     *              
     * @param  s String to be converted      
     * @return boolean conversion of String          
     */
    protected static boolean atob(String s) {return (s.equals("true")); }
   
    /**
     * Gets the simulation description
     *              
     * @return Simulation description, top level tree of objects  
     */
    public Object getDesc() { return desc; } 
    
    
    
    //JAG: This one was not there before
    
    /** Convert a relative URL to a full one */
    String toFullURL(String s) {
	if (s.startsWith("http:") || s.startsWith("ftp:") || s.startsWith("file:")) {	
	    return s;
	} else {
	    return "file:"+cwd + "/" + s;
	}
    }
    
    //JAG: Different before

    /**
     * Extracts the list of schema urls from the <module>  and top level nmlschema= attribute
     * so we can then load them in the classloader path.
     * @returns a vector of URLs
     */
    public Vector<URL> extractURL(org.jdom.Element root) {
	Vector<URL> plv = new Vector<URL>();
	neuroml.NeuromlPkg nmlpkg;
	org.jdom.Element modulesElement;

	// Look for a schema location... as a nmlschema="http://www.xxx.org/sch.jar"
	// Look for a schema location... as a nmlschema="http://www.xxx.org/sch.xsd"
	// Look for a schema location... as a nmlschema="http://www.xxx.org/sch.xml"
	// Look for a schema location... as a nmlschema="http://www.xxx.org/sch.xml;http://www.yyy.org/sch.xml"
	Attribute a = root.getAttribute("nmlschema");
	nmlschema=null;

	if (a!=null) {
	    nmlschema = a.getValue();
	    // The schema can be in a ; separated list...
	    StringTokenizer st = new StringTokenizer(nmlschema, ";,\n\r\t");
	    while (st.hasMoreTokens()) {
		String nt = st.nextToken();
		nt = toFullURL( nt );		
		try {
		    plv.addElement( new URL(nt) );
		} catch (Exception e) {
		    System.out.println(e +" faulty url "+nt);
		}
	    }
	}

	if (root.getName().equals("neuroml")) {
	    modulesElement = root.getChild("modules");
	}  else {
	    // Assume root name is a class name...
	    // Don't look for a <modules> section
	    return plv;
	}
	
	// Build the NeuromlPkg object without its subelements
	String className = getFieldClass(root,null,false);
	Object obj = NeuromlSys.makeNeuromlObject(className);
	if (obj instanceof neuroml.NeuromlPkg)
	    nmlpkg = (neuroml.NeuromlPkg)obj;
	else {
	    return plv;
	}

	// Build only the <modules> subelement
	Set ms = (Set)startElement(modulesElement,nmlpkg);
	Vector mv = ms.getElements();
	try {
	    for (int i=0;i<mv.size();i++) {
		Module mod = (Module)mv.get(i);
		
		plv.add( new URL(toFullURL(mod.location)) );
	    }
	} catch (MalformedURLException eurl) {
	    System.out.println("NeuroMLParser extractURL : Malformed URL "+eurl);
	}
	return plv;
    }

    /** Try to create a class.
     * Look in: jar file, std Class loader, list of known package prefixes
     */
    private Class createClass(String className) {
	Class c = null;
	try { c = NeuromlSys.classLoader.loadClassFromJar(className); } catch (Exception e) {}
	if (c == null) try { c = Class.forName(className); } catch (Exception e) {}
	if (c == null) {
	    /***** This is a slow algorithm... 
	     ***** Search from top, should warn about conflicts...
	     *****/
	    Vector kpl = XMLIn.getKnownPackagesList();
	    for (int j=kpl.size()-1;j>=0;j--) {
		String packageName = (String)kpl.elementAt(j);
		try { c = Class.forName(packageName+"."+className); } catch (Exception e) {}
		/**** Should we also look in the .jar file for the short name??? ****/
		if (c!=null) break;
	    }
	}
	return c;
    }    
    
    /** Tries to create an object of the given class using Class.forName
     */
    private Object createObject(String className) {
 	Class c = createClass(className);   
	Object o = null;
	if (c!=null) {
	    try { o = c.newInstance(); } 
            catch (Exception e) 
            {
                System.out.println("Something strange with "+c.getName());
            }
	}
	return o;
    }
    
    /** Set the instance name if object is Nameable
     */
    public void setInstName(org.jdom.Element element, Object instElement) {
	if (instElement instanceof Nameable) {
	    ((Nameable)instElement).setName( getFieldName(element) );
	}
    }
 

    /** Copy across the attributes of <element attr="" into  the object.
     */
    public void setAttributes(org.jdom.Element element, Object ie) {
	java.util.List attributesList = element.getAttributes();

	Class cd = ie.getClass();
	Field cf;

	for (int i = 0; i < attributesList.size(); i++) { 
	    Attribute attr = (Attribute)attributesList.get(i);
	    String fn = attr.getName(); // Attribute Name=
 
	    if ((!fn.equals("class")) && (!fn.equals("name"))) {  
		String fv = attr.getValue(); // name="123"
		try {
		    cf = cd.getField(fn);  // name of field in class { int field; }
		    String ft = cf.getType().getName();
		    String sft = getShortClassName( ft );
		    // Set the value in the field appropriately...
		    if (sft.equals("int")) { 
			cf.setInt(ie, atoi(fv));
		    } else if (sft.equals("short")) { 
			cf.setShort(ie, atos(fv));
		    } else if (sft.equals("boolean")) { 
			cf.setBoolean(ie, atob(fv));
		    } else if (sft.equals("double")) { 
			cf.setDouble(ie, atod(fv));
		    } else if (sft.equals("float")) { 
			cf.setFloat(ie, atof(fv));
		    } else if (sft.equals("String")) {
			cf.set(ie, fv);
                        //JAG module treat
                         if(element.getName().compareTo("module")==0) 
                        {
                            JarClassLoader truc=NeuromlSys.classLoader;
                            if(truc!=null)
                            {
                                String[] strings=truc.readJar(fv);
                                ns.initialiseDerivedFrom(strings);
                            }
                            else System.out.println(element.getName()+" is null!");
                        }
 		    } else if (sft.equals("Integer")) {
			cf.set(ie, new Integer(fv));
                    } else if (sft.equals("Short")) {
			cf.set(ie, new Short(fv));                        
 		    } else if (sft.equals("Double")) {
			cf.set(ie, new Double(fv));
 		    } else if (sft.equals("Float")) {
			cf.set(ie, new Float(fv));
 		    } else if (sft.equals("Boolean")) {
			cf.set(ie, new Boolean(fv));
			// 		    } else if (ft.equals("Ref")) {		
			//			Ref ref = startRef(fn,fv,ie);
			//			cf.set(ie, ref);
		    } else if (sft.equals("DValue")) {		
			DValue dvalue = getDValue(fn,fv);
			cf.set(ie, dvalue);
		    } else {
			// Try making a new object for the attribute...
			Object fieldObj = null;
			try { fieldObj = cf.getType().newInstance(); } catch (Exception e) {}
			
			if (fieldObj instanceof neuroml.core.Enum) {
			    // System.out.println("FWH setParameter Enum "+ft);
			    neuroml.core.Enum e = (neuroml.core.Enum) ( fieldObj );
			    e.setValue( fv );
			    cf.set(ie, e);
			} else if (fieldObj instanceof neuroml.core.Reference) {
			    Reference ref = startRef(fn,fv,ie);
			    cf.set(ie, ref);
			} else {
			    System.out.println("Error in set Attrs =ft "+ft+" not handled");
			}	
		    } 	   		    		    
		} catch (Exception e) {
		    // Ignore problems if schema location specified.
		    // We're not using it yet.
		    if (!fn.equals("schemaLocation")) {
			if (!fn.equals("nmlschema")) {
			    System.out.println("Warning - error in JDOMNeuroMLParser.setAttributes finding"+
					       "field named "+ fn+" in class "+cd.getName() +
					       "element "+element+" ("+ e+")");
			}
		    }
		} 
	    }
	}      
    }
    
    /** Tries to get the attribute "name"= out of a <element>
     */
    public String getFieldName(org.jdom.Element element){
	Attribute a = element.getAttribute("name");
	if (a!=null) { 
	    return a.getValue(); 
	} else {
	    return element.getName();
	}
    }

    /** Gets a given field, or null if not present
     */
    public String getFieldParameter(org.jdom.Element element, String name){	
	Attribute a =  element.getAttribute(name);
	if (a!=null) { return a.getValue(); }
	return null;
    }

    /** Gets the class for this element.
     * @param listed : true if the element is part of a  set 
     */
    public String getFieldClass(org.jdom.Element element, Object parent, boolean listed) {
	String cn = getFieldParameter(element,"class");
	if (cn!=null) { 
	    return cn;
	} else {
	    if (parent != null) {		
		Class cd = parent.getClass();

		if (listed) { // element of a set/list
		    if (XMLIn.getUseTypeNameInSet()) {
			// If flag set, use set element names as the type
			return element.getName();
		    } else {
			if (parent instanceof ContentRestricted) {
			    return ((ContentRestricted)parent).getContentClassName();
			} else {
			    return "Object";
			}					 
		    }
		} else {
		    Field cf = getFieldNamed(element.getName(),parent);		    
		    if (cf != null) {
			return cf.getType().getName();
		    } else {
			System.out.println("Neuroml Parser Warning : "+element.getName()+
					   " not found in schema.  Maybe the class name "+
					   parent.getClass().getName()+
					   " is not unique?"); 
			return null;
		    }
		}
	    } else {
		// We're the root element, but no class= was specified 
		// Assume the format is <ClassName xmlns.../>
		return element.getName();	 
	    }
	}	
    }

    /** Gets the field of this name if it exists in the class or its superclasses
     */
    public Field getFieldNamed(String fn, Object parent) {
	Class cd = parent.getClass();
	Field cf = null;
	try {
	    cf = cd.getField(fn);
	    if (debug) System.out.println("get field named "+fn+" in class "+
					  cd.getName()+" is "+cf);
	} catch (NoSuchFieldException e) {

	    // if the tag name of the element is not recognized as a 
	    // field of the parent, we check if this tag name is a field of the parent superclasses	    
	    Class superclass = cd.getSuperclass();	    
	    if (debug) System.out.println("set field name : no such field "+fn+
					  " in class "+cd.getName());
	    while (superclass != null) {
		//String cn = getShortClassName(superclass.getName());
		try {
		    if (debug)
			System.out.println("try field name "+fn+
					   " in class = "+superclass.getName());
		    cf = superclass.getField(fn);
		    return cf;		    
		} catch (NoSuchFieldException e2) {
		}		
		superclass = superclass.getSuperclass();
	    }
	}
	return cf;
    }

    /** Sets a parameter
     */
    public void setParameter(String fn, Object subInstElement, Object ie) {
	Field cf;	
	if (debug) System.out.println("Set paramater: field name = "+fn);
	try {
	    cf = getFieldNamed(fn,ie);
	    if (cf==null) {
		if (debug) System.out.println("set field name : no such field "+fn);	   
	    } else {
		if (debug) System.out.println("field name : "+fn+" type : "+
					      cf+" set object type "+subInstElement);	   
		cf.set(ie, subInstElement);
	    }
	} catch (IllegalAccessException e) {
	    System.out.println("field "+fn+" "+e);
	}	
    }

    /** Checks if a field is a basic type
     */
    public boolean isBasicType(String ft) {
	String s = getShortClassName(ft);
	return (s.equals("int") || s.equals("short") || s.equals("bool") ||
		 s.equals("double") || s.equals("float") || s.equals("String"));
    }

    /** Check if a field is a basic type when in a Set or a List (Object versions)
     */
    public boolean isBasicTypeInList(String ft) {
	String s = getShortClassName(ft);
	return (s.equals("Integer") ||s.equals("Short") || s.equals("Boolean") ||
		s.equals("Double") || s.equals("Float") || s.equals("String"));
    }

    /** Check if field is an array
     */
    public boolean isArray(String ft) {
	String s = getShortClassName(ft);
	if ((s.equals("DoubleArray"))||(s.equals("IntArray"))||
	   (s.equals("StringArray")))
	    return true;
	else
	    return false;	
    }

    /** Extracts whitespace separated values from an array
     */
    public String[] extractValues(String content, int length) {
	String[] values = new String[length];
	String value;
	int index,preindex;
	preindex = 0;
	if (debug) { 
	    System.out.println("extractValues "+content+" "+length); 
	}
	for (int i=0;i<length;i++) {
	    index = content.indexOf(' ', preindex);
	    if (index==-1) { // no more space in the string
		if (debug) { System.out.println("Last Index "+preindex+" "+length); }
		values[i] = content.substring(preindex,content.length());
		break;
	    } else {
		values[i] = content.substring(preindex,index);
		if (debug) { System.out.println(" got "+i+ " "+values[i]); }	
		preindex = index+1;
	    }
	}
	return values;
    }

    /** Create an array object
     */
    
    //Different from JAG GUI version
    public Object getArray(org.jdom.Element element, Class expClass) {
	int length = atoi(getFieldParameter(element,"length"));
	String[] values;
	values = extractValues(element.getTextTrim(),length);
	
	if (expClass.isAssignableFrom(DoubleArray.class)) {	    
	    DoubleArray array = new DoubleArray();
	    double[] data = new double[length];
	    for(int i=0;i<length;i++){
		data[i] = atod(values[i]);	
	    }
	    array.getUnitRef().setTargetName(getFieldParameter(element,"unit"));
	    array.setData(data);
	    array.setName(getFieldName(element));
	    return array;
	} else if (expClass.isAssignableFrom(IntArray.class)) {	    
	    IntArray array = new IntArray();
	    int[] data = new int[length];
	    for(int i=0;i<length;i++){
		data[i] = atoi(values[i]);	
	    }	    
	    array.setData(data);
	    array.setName(getFieldName(element));
	    return array;
	} else if (expClass.isAssignableFrom(StringArray.class)) {	 
	    StringArray array = new StringArray();
	    array.setData(values);
	    array.setName(getFieldName(element));
	    return array;
	} else {
	    System.out.println("NeuroMLParser Error startArray :"+
			       "type mismatch : name = "+getFieldName(element));
	    return null;
	}	
    }

    /** Returns true if character is a-z
     */
    private boolean charIsAlpha(char c) {
	return (((c>='a')&&(c<='z'))||((c>='A')&&(c<='Z')));
    }

    /** Returns true if character is a-z
     */
    private boolean charIsNumeric(char c) {
	return (c>='0' && c<='9') || c=='.' || c=='-' || c=='+';
    }
   
    /** Find the begining index of the unit string in a string containing the value and the unit 
     */
    private int findUnitIndex(String v) {
	int i=0;
	boolean found = false;
	int l = v.length();
	for (i=0; i<l && !found; i++) {
	    char c = v.charAt(i);
	    // Check for e
	    if (c=='e' || c=='E') {
		if (i+1<l) {
		    if (charIsNumeric(v.charAt(i+1))) {
			i++; // skip the e
		    }
		}
	    } 
	    // Normal check
	    found = ! charIsNumeric(v.charAt(i));
	}
	return i-1;
    }

    /** Creates a DValue
     */
    public DValue getDValue(String nameAsField, String valueAndUnit) {
	String valueAndUnitTrimmed = valueAndUnit.trim();
	DValue dvalue = new DValue();

	int unitIndex = findUnitIndex(valueAndUnitTrimmed);
	String unitName = valueAndUnitTrimmed.substring(unitIndex) ;
	String value = valueAndUnitTrimmed.substring(0,unitIndex) ;

	dvalue.getUnitRef().setTargetName(unitName);	  
	dvalue.setName(nameAsField);
	dvalue.setValue(atod(value));
	return dvalue;
    }
 
    /** Gets a Set from the Parent instantiation
     * ... so make sure the parent has instantiated it!!!
     * ... and this doesn't work if the parent is itself a set!!!
     */
    public java.util.Collection getSet(String setNameAsField, Object parent) {
	Field cf = getFieldNamed(setNameAsField,parent);
	if (cf == null) {
	    // maybe it's a set of sets...
	    //	    System.out.println("getSet "+setNameAsField+" thinks it got a setset");
	    return null;
	}
	Collection defaultSet = null;
	try {
	    defaultSet = (java.util.Collection)cf.get(parent); // get default set from parent	 
	} catch (Exception e) {
	    System.out.println(e);
	    e.printStackTrace();
	}
	return defaultSet ;
    } 

    /** Gets a Map (=Hash) from the Parent instantiation
     * ... so make sure the parent has instantiated it!!!
     * ... and this doesn't work if the parent is itself a set!!!
     */
    public java.util.Map getMap(String mapNameAsField, Object parent) {
	Field cf = getFieldNamed(mapNameAsField,parent);
	if (cf == null) {
	    return null;
	}
	Map defaultMap = null;
	try {
	    defaultMap = (java.util.Map)cf.get(parent); 
	} catch (Exception e) {
	    System.out.println(e);
	    e.printStackTrace();
	}
	return defaultMap;
    }        

    /** Creates a Reference
     */
    public Reference startRef(String refNameAsField, String target, Object parent) {
	Field cf = getFieldNamed(refNameAsField,parent);
	Reference defaultRef = null;
	try {
	    defaultRef = (Reference)cf.get(parent); // get default ref from parent...
	    defaultRef.setTargetName(target);
	    if (debug) System.out.println("# Ref "+ refNameAsField+" target : "+target);	
	} catch (IllegalAccessException e) {
	    System.out.println(e);
	}
	return defaultRef;
    }

    /** Creates the NeuroML object corresponding to a org.jdom.Element 
     * (an org.jdom.Element describes an xml tag)
     * @return the instantiated NeuroML object corresponding to this tag
     * @param element : the object desribing the xml tag
     * @param instParent : the NeuroML object created from the up level tag preceding this tag
     * @param type : "child" for element of list or set, anything if not
     */
    public Object startElement(org.jdom.Element element, Object instParent)  {
	return startElement(element,instParent,"element");
    }

    /** Create the NeuroML object corresponding to a org.jdom.Element 
     * Called for every <tag> in the XML file.
     *
     * @param element the JDOM version of a <tag>
     * @param instParent the parent object instance
     * @param type the type of the element.
     * @return the constructed object.
     */
    public Object startElement(org.jdom.Element element, Object instParent, String type) {
	// ie : NeuroML object to be returned by this function
	Object ie=null; 

	boolean childrenAreElementofSet = false;
	boolean childrenAreElementofMap = false;
	boolean listed = type.equals("child");
	boolean isBasicType = false;

	String elementTagName = element.getName(); // <thetag>
	String cn = getFieldClass(element,instParent,listed); // The expected class of field (from schema)

	// Try to make a Class object for the field
	Class expClass = null;
	boolean isDerivedFromEnum = false;
	boolean isDerivedFromRef = false;
	expClass = createClass( cn ); 
	if (expClass == null) { expClass = Object.class; }

	// Check if class is derived from Enum...
	isDerivedFromEnum = neuroml.core.Enum.class.isAssignableFrom(expClass);
	isDerivedFromRef = Reference.class.isAssignableFrom(expClass);

	String scn = getShortClassName( cn );
 
	if (cn==null) {
	    // do nothing - there's a problem
	} else if (isArray(cn)) {
	    if (debug) System.out.println("element tag is Array");
            //JAG modified something in version for GUI
	    ie = getArray(element, expClass); // Get a special array object...
	    return ie;	   
	} else if (java.util.Map.class.isAssignableFrom( expClass ) ){
	    ie = getMap(element.getName(),instParent);
	    if (ie == null) {
		// Make a set...
		try {
		    ie = (java.util.Map)expClass.newInstance();
		} catch (Exception e) {}
	    }
	    childrenAreElementofMap = true;
	    // Fill out children below...
	} else if (java.util.Collection.class.isAssignableFrom( expClass ) ){
            //JAG Flag
            if(element.getName().compareTo("NeuronsPosition")==0)
                    element=element;
            
	    ie = getSet(element.getName(),instParent);
	    if (ie == null) {
		// Make a set...
		try {
		    ie = (java.util.Collection)expClass.newInstance();
		} catch (Exception e) {}
	    }
	    childrenAreElementofSet = true;
            
            if (ie!=null && !isBasicType) {		
		setInstName(element,ie); // used for "name" because it's a private field
		setAttributes(element,ie);
	    }
	    // Fill out children below...
	} else if (isDerivedFromEnum) {
	    try { 
		neuroml.core.Enum e = (neuroml.core.Enum)expClass.newInstance();
		String txt = element.getTextTrim();
		if (txt==null || txt.equals("")) {
		    txt = getFieldParameter(element,"value");
		}
		e.setValue( txt );
		ie = e;
	    } catch (Exception e) {}
	} else if (isDerivedFromRef) {
	    try { 
		Reference r = (Reference)expClass.newInstance();
		String txt = element.getTextTrim();
		if (txt==null || txt.equals("")) {
		    txt = getFieldParameter(element,"value");
		}
		r.setTargetName( txt );
		ie = r;
	    } catch (Exception e) {}
	} else if (isBasicType(cn) || isBasicTypeInList(cn)) {
	    /** For Basic Types, allow either attributes or elements... 
	     * Here, assuming that they are elements...
	     */
	    // ... but if they are in a set, we get:
	    // <Integer value="123"/> not <Integer>3</Integer>
	    String txt = element.getTextTrim();
	    if (txt==null || txt.equals("")) {
		txt = getFieldParameter(element,"value");
	    }
	    try {
		if (scn.equals("Double") || scn.equals("double")){
		    ie = new Double(txt);
		} else if (scn.equals("Float") || scn.equals("float")) {
		    ie = new Float(txt);
		} else if (scn.equals("Integer") || scn.equals("int")) {
		    ie = new Integer(txt);
                } else if (scn.equals("Short") || scn.equals("short")) {
		    ie = new Short(txt);                    
		} else if (scn.equals("Boolean") || scn.equals("bool")) {
		    ie = new Boolean(txt);
		} else if (scn.equals("String")) {
		    ie = new String(element.getText());
		} else {
		    // Could be an Enum too... checked above using isDerivedFromEnum		    
		}
	    } catch (Exception e) {
		System.out.println("Error setting "+scn+" in "+element+" to "+txt+" "+e);
	    }
	} else { // a Class
	    String packageName = new String();
 	    if (!listed) {
		if (instParent!=null) {
		    Field cf = getFieldNamed(element.getName(),instParent);
		    Object defaultObj = null;
		    try {
			ie=(Object)cf.get(instParent); // get default object from parent
			if (!ie.getClass().getName().endsWith(cn)) {
			    // If default object not same class as specified,
			    // force making a new one...
			    ie = null;
			}
		    } catch (Exception e) {
			//System.out.println(e);
		    }
		}
	    } else {
		if (debug) System.out.println("We are in a set, don't try to poke parent's fields");
	    }

	    if (ie==null) {
		// try first to create basic type's object as they could be in set or list
		if (isBasicTypeInList(cn)) {		 
		    if (scn.equals("Double")) {
			String val = getFieldParameter(element,"value");
			ie = new Double(val);
		    } else if (scn.equals("Integer")) {
			String val = getFieldParameter(element,"value");
			ie = new Integer(val);
                    } else if (scn.equals("Short")) {
			String val = getFieldParameter(element,"value");
			ie = new Short(val);                        
		    } else if (scn.equals("Float")) {
			String val = getFieldParameter(element,"value");
			ie = new Float(val);
		    } else if (scn.equals("Boolean")) {
			String val = getFieldParameter(element,"value");
			ie = new Boolean(val);
		    } else if (scn.equals("String")) {
			String val = getFieldParameter(element,"value");
			ie = new String(val);
		    } 
		    if (ie!=null) isBasicType = true;
		} else {
		    // instantiate the object, searching for the full name of its class
		    ie = createObject(cn);	 
		    if (ie==null) {
			System.out.println("\n\n***** SERIOUS PROBLEM !!!! create class name = "+
					   packageName+"."+cn+
					   " failed!");
			System.out.println("Probably you forgot to have a public null constructor");
			System.out.println("   e.g. public "+cn+"() {}\n\n");
			System.out.println("Or maybe you didn't call XMLIn.addKnownPackage");
			System.out.println("to add the package name\n\n");
			return null;
		    }
		}
	    }

	    if (ie!=null && !isBasicType) {		
		setInstName(element,ie); // used for "name" because it's a private field
		setAttributes(element,ie);
	    }
	}

	// safety check
	if (ie==null) return null;

	// deal with children 
	Object subInstElement;
	java.util.List list = element.getChildren();
	org.jdom.Element subElement=null;
	
	for (int i=0; i<list.size(); i++) {
	    subElement = (org.jdom.Element)list.get(i);
	    if (childrenAreElementofSet) {
		subInstElement = startElement(subElement,ie,"child");
		if (subInstElement!=null) ((java.util.Collection)ie).add(subInstElement);
	    } else if (childrenAreElementofMap) {
		subInstElement = startElement(subElement,ie,"child");
		String key = getFieldParameter(subElement, "name");
		// Key is name field...
		if (subInstElement!=null && key!=null) {
		    ((java.util.Map)ie).put(key, subInstElement);
		}
	    } else {
		subInstElement = startElement(subElement,ie);
		if (subInstElement!=null) setParameter(getFieldName(subElement),subInstElement,ie);
	    }	    	    
	}	
	return ie;
    }
       
    /** Gives the shortclass name, the last name after the last '.'
     */
    private String getShortClassName(String cn) {
	if (cn == null) { return null; }
	int index = cn.lastIndexOf('.');
	if (index!=-1)
	    return cn.substring(index+1);
	else 
	    return cn;
    }  
}
