package neuroml.util;


import neuroml.core.*;
import neuroml.model.channel.*;
import neuroml.model.cell.*;
import neuroml.model.network.*;

import org.jdom.*;
import org.jdom.input.*;

import java.io.*;

import java.io.InputStream;
import java.io.FileInputStream;
import java.io.File;
import java.net.*;

import java.util.Hashtable;
import java.util.Vector;

/**
 * A class for reading in XML into an object tree.
 *
 * @author Paul Rogister      
 *
 * @version 1.0, 02 March 2001   
 *                 
 */
public class XMLIn {

    static NeuroMLParser domparser;

    /** Store static namespace - ** WARNING ** this 
     * gets changed whenever a document is loaded.
     */
    static String lastnamespace="global";
    
    private static boolean debug = false;

    static Vector<String> packagesList = null;

    static Hashtable<String,String> packageNames = new Hashtable<String,String>();

    /** Returns the list of known packages
     * - all the neuroml.ones, java.lang
     */
    public static Vector<String> getKnownPackagesList() { 
	if (packagesList == null) {
	    packagesList = new Vector<String>();
	    // Add neuroml.*...  perhaps this shouldn't be in the core neuroml
	    String[] s = neuroml.NeuromlPkg.getPackageList();
	    for (int i=0; i<s.length; i++) {
		packagesList.addElement( s[i] );
		packageNames.put( s[i], "ok" );
	    }	  
	    packagesList.addElement("java.lang");
	    packageNames.put( "java.lang", "ok" );		
	}
	return packagesList; 
    }

    /** Add a user package to the list. Used by parser to 
     * instantiate objects of user's classes using short name "MyClass" 
     * instead of the full path ("org.mysimulator.MyClass")
     * @param pname name of user package, e.g. org.mysimulator
     */
    public static void addKnownPackage(String pname) { 
	if (packagesList == null) { packagesList = getKnownPackagesList(); }
	if (!packageNames.containsKey(pname)) {
	    packagesList.addElement(pname);
	    packageNames.put(pname,"ok");
	}
    }

    /** A Field to choose whether to use field or type name in set elements
     * <pre>
     * false:- Set things = new Set("Widget"); -->
     * &lt;things>
     *    &lt;thing />
     *    &lt;thing class="MyWidget"/>
     * &lt;/things>
     *
     * true:-
     * &lt;things>
     *    &lt;Widget/>
     *    &lt;MyWidget/>
     * &lt;/things>
     * </pre>
     */
    static boolean useTypeNameInSet = false;

    public static boolean getUseTypeNameInSet() { return useTypeNameInSet; }
    public static void  setUseTypeNameInSet(boolean f) { useTypeNameInSet = f; }

    /** A Field to choose whether to use field or type name in set elements
     * <pre>
     * false:- Set things = new Set("Widget"); -->
     *    &lt;thing />
     *    &lt;thing class="MyWidget"/>
     *
     * true:-
     * &lt;things>
     *    &lt;Widget/>
     *    &lt;MyWidget/>
     * &lt;/things>
     * </pre>
     */
    static boolean useSetName = true;
    public static boolean getUseSetName() { return useSetName; }
    public static void  setUseSetName(boolean f) { useSetName = f; }

    /** Returns the last known top level schema location
     * <pre>&lt;neuroml class="a.b.c" nmlschema="file:./moduleCache/myplace.jar"/></pre>
     * @return the schema location
     */
    public static String getNmlSchema() { return domparser.getNmlSchema(); }


    /**
     * Convert a URL of an XML file into an object tree
     *
     * @param xmlfilename Name of the xml description file     
     * @return Top level object
     */
    public static Object loadXML( String xmlfilename ) {
	try {	    	    
	    // Try and get an input stream from the xmlfilename.
	    InputStream is;
	    if (xmlfilename.startsWith("http:") || xmlfilename.startsWith("ftp:") || 
		xmlfilename.startsWith("file:")) {
		try {
		    is = (new URL(xmlfilename)).openStream();
		} catch (Exception e) {
		    if (debug) System.out.println("XMLIn url exception "+e);
		    return null;
		}
	    } else {
		try {
		    is = new FileInputStream( new File( xmlfilename ) );
		} catch (Exception e) {
		    if (debug) System.out.println("XMLIn file exception "+e);
		    return null;
		}
	    }
	    lastnamespace = xmlfilename;

	    // read XML from an input stream
	    return loadXML( is );

        } catch (Throwable t) {
            t.printStackTrace ();
        }
	return null;
    }

    /**
     * Convert an input stream into an object tree
     *
     * @param xmlfilename Name of the xml description file     
     * @return Top level object
     */
    public static Object loadXML( InputStream is ) {
	try {	    	   
	    SAXBuilder parser = new SAXBuilder();
	    Document jdomDocument; 
	    jdomDocument = parser.build(is);	    
	    domparser = new NeuroMLParser(jdomDocument, lastnamespace);
	    return domparser.getDesc();
	} catch (Throwable t) {
            t.printStackTrace ();
        }
	return null;
    }

    /** Converts an XML string to an object
     * using default namespace
     */
    public static Object fromString( String xmltext ) {
	return fromString( xmltext, lastnamespace );
    }

    /** Converts an XML string to an object
     */
    public static Object fromString(String xmltext, String namespace) {
	try {
	    SAXBuilder parser = new SAXBuilder();
	    Document jdomDocument; 
	    jdomDocument = parser.build(new StringReader(xmltext));
	    domparser = new NeuroMLParser(jdomDocument, namespace);
	    lastnamespace = namespace;
	    return domparser.getDesc();
        } catch (Throwable t) {
            t.printStackTrace ();
        }
	return null;	
    }

    /** Retrieves an Object from a Reference
     * @param ref the Reference to deref
     * @return the object, or null
     */
    public static Object deRef(Reference ref) {
	return NeuromlSys.deRef( ref, lastnamespace );
    } 

    public static void addRef(String key, Object obj) {
	NeuromlSys.addRef( key, obj, lastnamespace );
    }

    public static void removeRef(String key, Object obj) {
	NeuromlSys.removeRef( key, obj, lastnamespace );
    }

    public static Hashtable getReferences() {
	Namespace ns = NeuromlSys.getNamespace(lastnamespace);
	if (ns!=null) { return ns.getReferences(); } else return null;
    }
}
